(******************************************************************************)
(***              SubCls - TWinControl subclass helper                      ***)
(******************************************************************************)
unit SubCls;

interface

uses
  Classes, Controls, Windows, Messages, SysUtils;

{ TControlSubclass is an abstract ancestor for objects that implement a message
  handling method for a control. To handle methods for a control in a component
  you should create a new class, derived from TControlSubclass, and handle the
  messages in the ...WndProc functions. }
type
  TControlSubclass = class
  protected
    procedure BeforeWndProc(Control: TControl; var Message: TMessage); virtual;
    procedure AfterWndProc(Control: TControl; var Message: TMessage); virtual;
  end;

procedure RegisterSubclass(Control: TControl; Subclass: TControlSubclass);
procedure UnRegisterSubclass(Control: TControl; Subclass: TControlSubclass);

implementation

{ TControlSubclass }

procedure TControlSubclass.BeforeWndProc(Control: TControl; var Message: TMessage);
begin
end;

procedure TControlSubclass.AfterWndProc(Control: TControl; var Message: TMessage);
begin
end;

{ TControlSubclassInfo }

type
  TControlSubclassInfo = class(TList)
  private
    procedure WndProc(var Message: TMessage);
  public
    Control: TControl;
    OriginalWindowProc: TWndMethod;
    procedure Add(ControlSubclass: TControlSubclass);
    procedure Remove(ControlSubclass: TControlSubclass);
  end;

procedure TControlSubclassInfo.WndProc(var Message: TMessage);
var
  i: Integer;
begin
  if Count > 0 then
    for i := 0 to Count - 1 do
      if Assigned(Items[i]) then
         TControlSubclass(Items[i]).BeforeWndProc(Control, Message);

  if Assigned(OriginalWindowProc) then
    OriginalWindowProc(Message);

  if Message.Msg <> CM_RELEASE then
  begin
    if Count > 0 then
      for i := 0 to Count - 1 do
        if Assigned(Items[i]) then
          TControlSubclass(Items[i]).AfterWndProc(Control, Message);
  end;
end;

procedure TControlSubclassInfo.Add(ControlSubclass: TControlSubclass);
begin
  Control.WindowProc := WndProc;
  inherited Add(ControlSubclass);
end;

procedure TControlSubclassInfo.Remove(ControlSubclass: TControlSubclass);
begin
  inherited Remove(ControlSubclass);
  if Count = 0 then
    Control.WindowProc := OriginalWindowProc;
end;

{ TControlSubclassList }

type
  TControlSubclassList = class(TList)
  public
    function FindControl(Control: TControl): Integer;
    procedure Add(Control: TControl; ControlSubclass: TControlSubclass);
    procedure Remove(Control: TControl; ControlSubclass: TControlSubclass);
  end;

var
  ControlSubclassList: TControlSubclassList;

function TControlSubclassList.FindControl(Control: TControl): Integer;
var
  i: Integer;
begin
  Result := -1;
  if Count > 0 then
    for i := 0 to Count-1 do
      if TControlSubclassInfo(Items[i]).Control = Control then
      begin
        Result := i;
        Break;
      end;
end;

procedure TControlSubclassList.Add(Control: TControl; ControlSubclass: TControlSubclass);
var
  i: Integer;
  ControlSubclassInfo: TControlSubclassInfo;
begin
  i := FindControl(Control);

  if i < 0 then
  begin
    ControlSubclassInfo := TControlSubclassInfo.Create;
    ControlSubclassInfo.Control := Control;
    ControlSubclassInfo.OriginalWindowProc := Control.WindowProc;
    inherited Add(ControlSubclassInfo);
  end
  else
    ControlSubclassInfo := TControlSubclassInfo(Items[i]);

  ControlSubclassInfo.Add(ControlSubclass);
end;

procedure TControlSubclassList.Remove(Control: TControl; ControlSubclass: TControlSubclass);
var
  i: Integer;
  ControlSubclassInfo: TControlSubclassInfo;
begin
  i := FindControl(Control);

  if i >= 0 then
  begin
    ControlSubclassInfo := TControlSubclassInfo(Items[i]);
    ControlSubclassInfo.Remove(ControlSubclass);

    if ControlSubclassInfo.Count = 0 then
    begin
      inherited Remove(ControlSubclassInfo);
      ControlSubclassInfo.Free;
    end;
  end;
end;

procedure RegisterSubclass(Control: TControl; Subclass: TControlSubclass);
begin
  ControlSubclassList.Add(Control, Subclass);
end;

procedure UnRegisterSubclass(Control: TControl; Subclass: TControlSubclass);
begin
  ControlSubclassList.Remove(Control, Subclass);
end;

initialization
  ControlSubclassList := TControlSubclassList.Create;
finalization
  ControlSubclassList.Free;
end.
